#include <windows.h>
#include <gl/gl.h>
#include <gl/glu.h>
#include <stdio.h>
#include <math.h>
#include "glati.h"
#include "tga.h"
#include "overlay.h"
#include "font.h"
#include "md2.h"
#include "resource.h"

HWND hWnd=NULL;
HWND hWndDialog=NULL;
HDC hDC=NULL;
HGLRC hRC=NULL;

char *szAppName="OpenGL";

RECT WindowRect;

int Width=800, Height=600;

BOOL Done=FALSE, Key[255];

LARGE_INTEGER timerFrequency;
LARGE_INTEGER startTime;
LARGE_INTEGER currentTime;
float fps;
int nFrames=0;
char OutBuffer[255];

unsigned int FontBase, FontTexture, Base0, Bump0, Base1, Bump1;

typedef struct
{
	POINT pt;
	int x, y;
	int oldx, oldy;
	int dx, dy;
} mouse_t;

mouse_t mouse;

float roll, pitch=90.0f, yaw, zoom, i;
float m[16], light0[3]={ 0.0f, 0.0f, 50.0f };

MD2_Model_t Cube, Quad, Triangle, Sphere;

PFNGLACTIVETEXTUREARBPROC glActiveTextureARB=NULL;
PFNGLMULTITEXCOORD2FARBPROC glMultiTexCoord2fARB=NULL;

LRESULT CALLBACK WndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK DialogProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
void Render(void);
BOOL Init(void);
BOOL Create(void);
void Destroy(void);
void Resize(int Width, int Height);
BOOL IsExtensionSupported(const char *extension);

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int iCmdShow)
{
	WNDCLASS wc;
	MSG msg;

	wc.style=CS_VREDRAW|CS_HREDRAW|CS_OWNDC;
	wc.lpfnWndProc=WndProc;
	wc.cbClsExtra=0;
	wc.cbWndExtra=0;
	wc.hInstance=hInstance;
	wc.hIcon=LoadIcon(hInstance, MAKEINTRESOURCE(IDI_MAIN));
	wc.hCursor=LoadCursor(NULL, IDC_ARROW);
	wc.hbrBackground=GetStockObject(WHITE_BRUSH);
	wc.lpszMenuName=NULL;
	wc.lpszClassName=szAppName;

	RegisterClass(&wc);

	WindowRect.left=0;
	WindowRect.right=Width;
	WindowRect.top=0;
	WindowRect.bottom=Height;

	AdjustWindowRect(&WindowRect, WS_OVERLAPPEDWINDOW, FALSE);

	hWnd=CreateWindow(szAppName, szAppName, WS_OVERLAPPEDWINDOW|WS_CLIPSIBLINGS, CW_USEDEFAULT, CW_USEDEFAULT, WindowRect.right-WindowRect.left, WindowRect.bottom-WindowRect.top, NULL, NULL, hInstance, NULL);

	ShowWindow(hWnd, SW_SHOW);

	hWndDialog=CreateDialog(hInstance, MAKEINTRESOURCE(IDD_CONTROLPANEL), NULL, DialogProc);

	SetForegroundWindow(hWnd);

	if(!Create())
		Done=TRUE;

	if(Done==FALSE)
		Resize(Width, Height);

	if(Done==FALSE)
		if(!Init())
			Done=TRUE;

	while(!Done)
	{
		if(PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
		{
			if(msg.message==WM_QUIT)
				Done=TRUE;
			else
			{
				TranslateMessage(&msg);
				DispatchMessage(&msg);
			}
		}
		else
		{
			Render();
			SwapBuffers(hDC);

			nFrames++;
			if(nFrames>20)
			{
				QueryPerformanceCounter(&currentTime);
				fps=(float)nFrames/((float)(currentTime.QuadPart - startTime.QuadPart)/(float)timerFrequency.QuadPart);
				sprintf(OutBuffer, "FPS: %0.1f", fps);
				nFrames=0;
				QueryPerformanceCounter(&startTime);
			}
		}
	}

	Destroy();

	DestroyWindow(hWndDialog);
	DestroyWindow(hWnd);

	return msg.wParam;
}

LRESULT CALLBACK WndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch(uMsg)
	{
		case WM_CREATE:
			QueryPerformanceFrequency(&timerFrequency);
			QueryPerformanceCounter(&startTime);
			break;

		case WM_CLOSE:
			PostQuitMessage(0);
			break;

		case WM_DESTROY:
			break;

		case WM_SIZE:
			Width=LOWORD(lParam);
			Height=HIWORD(lParam);
			Resize(Width, Height);
			break;

		case WM_LBUTTONDOWN:
		case WM_MBUTTONDOWN:
		case WM_RBUTTONDOWN:
			SetCapture(hWnd);
			GetCursorPos(&mouse.pt);
			ShowCursor(FALSE);
			mouse.x=mouse.pt.x;
			mouse.y=mouse.pt.y;
			mouse.oldx=mouse.x;
			mouse.oldy=mouse.y;
			break;

		case WM_LBUTTONUP:
		case WM_MBUTTONUP:
		case WM_RBUTTONUP:
			if(!(wParam&(MK_LBUTTON|MK_RBUTTON|MK_MBUTTON)))
			{
				ShowCursor(TRUE);
				ReleaseCapture();
			}
			break;

		case WM_MOUSEMOVE:
			GetCursorPos(&mouse.pt);
			mouse.x=mouse.pt.x;
			mouse.y=mouse.pt.y;

			if(!wParam)
			{
				mouse.oldx=mouse.x;
				mouse.oldy=mouse.y;
				break;
			}

			mouse.dx=mouse.x-mouse.oldx;
			mouse.dy=mouse.oldy-mouse.y;

			if(!mouse.dx&&!mouse.dy)
				break;

			SetCursorPos(mouse.oldx, mouse.oldy);

			if(wParam==(MK_LBUTTON|MK_CONTROL))
			{
				if(abs(mouse.dx)>abs(mouse.dy))
					roll-=mouse.dx;
				else
					roll-=mouse.dy;
			}

			if(wParam==MK_LBUTTON)
			{
				yaw+=mouse.dx;
				pitch+=mouse.dy;
			}

			if(wParam==MK_RBUTTON)
				zoom+=mouse.dy;
			break;

		case WM_KEYDOWN:
			Key[wParam]=TRUE;

			switch(wParam)
			{
				case VK_ESCAPE:
					PostQuitMessage(0);
					break;

				default:
					break;
			}
			break;

		case WM_KEYUP:
			Key[wParam]=FALSE;
			break;
	}

	return DefWindowProc(hWnd, uMsg, wParam, lParam);
}

LRESULT CALLBACK DialogProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch(uMsg)
	{
		case WM_INITDIALOG:
			CheckDlgButton(hWnd, IDC_BASEMAP, BST_CHECKED);
			CheckDlgButton(hWnd, IDC_BUMPMAP, BST_CHECKED);
			CheckDlgButton(hWnd, IDC_CUBE, BST_CHECKED);
			CheckDlgButton(hWnd, IDC_SET1, BST_CHECKED);
			break;
	}

	return FALSE;
}

void matrixInvert(float m[16])
{
	float temp[16];

	temp[0]=m[0];
	temp[1]=m[4];
	temp[2]=m[8];
	temp[3]=0.0f;
	temp[4]=m[1];
	temp[5]=m[5];
	temp[6]=m[9];
	temp[7]=0.0f;
	temp[8]=m[2];
	temp[9]=m[6];
	temp[10]=m[10];
	temp[11]=0.0f;
	temp[15]=1.0f;
	temp[12]=-(m[12]*m[0])-(m[13]*m[1])-(m[14]*m[2]);
	temp[13]=-(m[12]*m[4])-(m[13]*m[5])-(m[14]*m[6]);
	temp[14]=-(m[12]*m[8])-(m[13]*m[9])-(m[14]*m[10]);

	memcpy(m, temp, sizeof(float)*16);
}

void Render(void)
{
	float lv0[3];

	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

	glPushMatrix();
	glTranslatef(0.0f, 0.0f, (zoom/16)-50.0f);
	glRotatef(yaw, 0.0f, 1.0f, 0.0f);
	glRotatef(-pitch, 1.0f, 0.0f, 0.0f);
	glRotatef(-roll, 0.0f, 0.0f, 1.0f);
	glGetFloatv(GL_MODELVIEW_MATRIX, m);
	matrixInvert(m);

	lv0[0]=(light0[0]*m[0])+(light0[1]*m[4])+(light0[2]*m[8]);
	lv0[1]=(light0[0]*m[1])+(light0[1]*m[5])+(light0[2]*m[9]);
	lv0[2]=(light0[0]*m[2])+(light0[1]*m[6])+(light0[2]*m[10]);

	glCullFace(GL_FRONT);

	glActiveTextureARB(GL_TEXTURE0_ARB);

	if(IsDlgButtonChecked(hWndDialog, IDC_SET1)==BST_CHECKED)
		glBindTexture(GL_TEXTURE_2D, Bump0);

	if(IsDlgButtonChecked(hWndDialog, IDC_SET2)==BST_CHECKED)
		glBindTexture(GL_TEXTURE_2D, Bump1);

	if(IsDlgButtonChecked(hWndDialog, IDC_BUMPMAP)==BST_CHECKED)
		glEnable(GL_TEXTURE_2D);
	else
		glDisable(GL_TEXTURE_2D);
	
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE_EXT);
	glTexEnvf(GL_TEXTURE_ENV, GL_COMBINE_RGB_EXT, GL_DOT3_RGB_EXT);
	glTexEnvf(GL_TEXTURE_ENV, GL_SOURCE0_RGB_EXT, GL_PRIMARY_COLOR_EXT);
	glTexEnvf(GL_TEXTURE_ENV, GL_OPERAND0_RGB_EXT, GL_SRC_COLOR);
	glTexEnvf(GL_TEXTURE_ENV, GL_SOURCE1_RGB_EXT, GL_TEXTURE);
	glTexEnvf(GL_TEXTURE_ENV, GL_OPERAND1_RGB_EXT, GL_SRC_COLOR);

	glActiveTextureARB(GL_TEXTURE1_ARB);

	if(IsDlgButtonChecked(hWndDialog, IDC_SET1)==BST_CHECKED)
		glBindTexture(GL_TEXTURE_2D, Base0);

	if(IsDlgButtonChecked(hWndDialog, IDC_SET2)==BST_CHECKED)
		glBindTexture(GL_TEXTURE_2D, Base1);

	if(IsDlgButtonChecked(hWndDialog, IDC_BASEMAP)==BST_CHECKED)
		glEnable(GL_TEXTURE_2D);
	else
		glDisable(GL_TEXTURE_2D);

	glPushMatrix();
		if(IsDlgButtonChecked(hWndDialog, IDC_CUBE)==BST_CHECKED)
			MD2_DrawFrameDOT3(&Cube, 0, lv0);

		if(IsDlgButtonChecked(hWndDialog, IDC_QUAD)==BST_CHECKED)
			MD2_DrawFrameDOT3(&Quad, 0, lv0);

		if(IsDlgButtonChecked(hWndDialog, IDC_TRIANGLE)==BST_CHECKED)
			MD2_DrawFrameDOT3(&Triangle, 0, lv0);

		if(IsDlgButtonChecked(hWndDialog, IDC_SPHERE)==BST_CHECKED)
			MD2_DrawFrameDOT3(&Sphere, 0, lv0);
	glPopMatrix();

	glActiveTextureARB(GL_TEXTURE1_ARB);
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	glDisable(GL_TEXTURE_2D);

	glActiveTextureARB(GL_TEXTURE0_ARB);
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	glDisable(GL_TEXTURE_2D);

	if(IsDlgButtonChecked(hWndDialog, IDC_SHOWVECTORS)==BST_CHECKED)
	{
		glPushMatrix();
			glScalef(1.001f, 1.001f, 1.001f);

			if(IsDlgButtonChecked(hWndDialog, IDC_CUBE)==BST_CHECKED)
				MD2_DrawVectors(&Cube, 0);

			if(IsDlgButtonChecked(hWndDialog, IDC_QUAD)==BST_CHECKED)
				MD2_DrawVectors(&Quad, 0);

			if(IsDlgButtonChecked(hWndDialog, IDC_TRIANGLE)==BST_CHECKED)
				MD2_DrawVectors(&Triangle, 0);

			if(IsDlgButtonChecked(hWndDialog, IDC_SPHERE)==BST_CHECKED)
				MD2_DrawVectors(&Sphere, 0);
		glPopMatrix();
	}

	glPopMatrix();

	glCullFace(GL_BACK);

	BeginOverlay(Width, Height);
		if(IsDlgButtonChecked(hWndDialog, IDC_SHOWVECTORS)==BST_CHECKED)
		{
			glColor3f(1.0f, 0.0f, 0.0f);
			Font_Print(0, 64, FontBase, FontTexture, "Red=Normal");
			glColor3f(0.0f, 1.0f, 0.0f);
			Font_Print(0, 48, FontBase, FontTexture, "Green=Tangent");
			glColor3f(0.0f, 0.0f, 1.0f);
			Font_Print(0, 32, FontBase, FontTexture, "Blue=Binormal");
		}

		glColor3f(1.0f, 1.0f, 1.0f);
		Font_Print(0, 16, FontBase, FontTexture, "%s", gluErrorString(glGetError()));
		Font_Print(0, 0, FontBase, FontTexture, OutBuffer);
	EndOverlay();
}

BOOL Init(void)
{
	if(!IsExtensionSupported("GL_ARB_multitexture"))
	{
		MessageBox(hWnd, "GL_ARB_multitexture not found!", "OpenGL", MB_OK);
		return 0;
	}
	else
	{
		glActiveTextureARB=(PFNGLACTIVETEXTUREARBPROC)wglGetProcAddress("glActiveTextureARB");
		glMultiTexCoord2fARB=(PFNGLMULTITEXCOORD2FARBPROC)wglGetProcAddress("glMultiTexCoord2fARB");
	}

	if(!IsExtensionSupported("GL_EXT_texture_env_combine"))
	{
		MessageBox(hWnd, "GL_EXT_texture_env_combine not found!", "OpenGL", MB_OK);
		return 0;
	}

	if(!IsExtensionSupported("GL_EXT_texture_env_dot3"))
	{
		MessageBox(hWnd, "GL_EXT_texture_env_dot3 not found!", "OpenGL", MB_OK);
		return 0;
	}

	FontBase=Font_Build();
	FontTexture=TGA_UploadImage("font.tga");

	Base0=TGA_UploadImage("base0.tga");
	Bump0=TGA_UploadDOT3("bump0.tga");
	Base1=TGA_UploadImage("base1.tga");
	Bump1=TGA_UploadDOT3("bump1.tga");

	MD2_Load("cube.md2", &Cube);
	MD2_Load("quad.md2", &Quad);
	MD2_Load("triangle.md2", &Triangle);
	MD2_Load("sphere.md2", &Sphere);

	glEnable(GL_DEPTH_TEST);
	glEnable(GL_CULL_FACE);

	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

	return TRUE;
}

BOOL Create(void)
{
	PIXELFORMATDESCRIPTOR pfd;
	unsigned int PixelFormat;

	pfd.nSize=sizeof(PIXELFORMATDESCRIPTOR);
	pfd.nVersion=1;
	pfd.dwFlags=PFD_DRAW_TO_WINDOW|PFD_SUPPORT_OPENGL|PFD_DOUBLEBUFFER;
	pfd.iPixelType=PFD_TYPE_RGBA;
	pfd.cColorBits=32;
	pfd.cDepthBits=24;

	hDC=GetDC(hWnd);

	if(!(PixelFormat=ChoosePixelFormat(hDC, &pfd)))
	{
		MessageBox(hWnd, "ChoosePixelFormat Failed.", "ERROR", MB_OK);
		return FALSE;
	}

	if(!SetPixelFormat(hDC, PixelFormat, &pfd))
	{
		MessageBox(hWnd, "SetPixelFormat Failed.", "ERROR", MB_OK);
		return FALSE;
	}

	if(!(hRC=wglCreateContext(hDC)))
	{
		MessageBox(hWnd, "wglCreateContext Failed.", "ERROR", MB_OK);
		return FALSE;
	}

	if(!wglMakeCurrent(hDC, hRC))
	{
		MessageBox(hWnd, "wglMakeCurrent Failed.", "ERROR", MB_OK);
		return FALSE;
	}

	return TRUE;
}

void Destroy(void)
{
	wglMakeCurrent(NULL, NULL);
	wglDeleteContext(hRC);
	ReleaseDC(hWnd, hDC);
}

void Resize(int Width, int Height)
{
	if(Height==0)
		Height=1;

	glViewport(0,0, Width, Height);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(90.0f, (float)Width/Height, 0.001f, 400.0f);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}

BOOL IsExtensionSupported(const char *extension)
{
	const unsigned char *extensions=NULL, *start, *where, *terminator;

	where=strchr(extension, ' ');

	if(where||*extension=='\0')
		return FALSE;

	extensions=glGetString(GL_EXTENSIONS);
	start=extensions;

	for(;;)
	{
		where=strstr(start, extension);

		if(!where)
			break;

		terminator=where+strlen(extension);

		if(where==start||*(where-1)==' ')
			if(*terminator==' '||*terminator=='\0')
				return TRUE;

		start=terminator;
	}

	return FALSE;
}
