#include <windows.h>
#include <gl/gl.h>
#include <gl/glu.h>
#include <stdio.h>
#include <malloc.h>
#include <math.h>
#include "glati.h"
#include "md2.h"

extern PFNGLMULTITEXCOORD2FARBPROC glMultiTexCoord2fARB;

void tangent_basis(float v0[3], float v1[3], float v2[3], float t0[2], float t1[2], float t2[2], float *tangent, float *binormal)
{
	float cp[3];
	float e0[3]={ v1[0]-v0[0], t1[0]-t0[0], t1[1]-t0[1] };
	float e1[3]={ v2[0]-v0[0], t2[0]-t0[0], t2[1]-t0[1] };

	cp[0]=e0[1]*e1[2]-e0[2]*e1[1];
	cp[1]=e0[2]*e1[0]-e0[0]*e1[2];
	cp[2]=e0[0]*e1[1]-e0[1]*e1[0];

	if(fabs(cp[0])>0.00001f)
	{
		tangent[0]=-cp[1]/cp[0];
		binormal[0]=-cp[2]/cp[0];
	}

	e0[0]=v1[1]-v0[1];
	e1[0]=v2[1]-v0[1];

	cp[0]=e0[1]*e1[2]-e0[2]*e1[1];
	cp[1]=e0[2]*e1[0]-e0[0]*e1[2];
	cp[2]=e0[0]*e1[1]-e0[1]*e1[0];

	if(fabs(cp[0])>0.00001f)
	{
		tangent[1]=-cp[1]/cp[0];
		binormal[1]=-cp[2]/cp[0];
	}

	e0[0]=v1[2]-v0[2];
	e1[0]=v2[2]-v0[2];

	cp[0]=e0[1]*e1[2]-e0[2]*e1[1];
	cp[1]=e0[2]*e1[0]-e0[0]*e1[2];
	cp[2]=e0[0]*e1[1]-e0[1]*e1[0];

	if(fabs(cp[0])>0.00001f)
	{
		tangent[2]=-cp[1]/cp[0];
		binormal[2]=-cp[2]/cp[0];
	}
}

BOOL MD2_Load(char *filename, MD2_Model_t *MD2)
{
	FILE *stream;
	int i, j, k;
	unsigned char buffer[MD2_MAX_FRAMESIZE];

	if((stream=fopen(filename, "rb"))==NULL)
		return FALSE;

	fread(&MD2->header, sizeof(unsigned char), sizeof(MD2_Header_t), stream);

	if(MD2->header.magic!=MD2_MAGIC||MD2->header.version!=MD2_VERSION)
	{
		fclose(stream);
		return FALSE;
	}

	fseek(stream, MD2->header.offsetSkins, SEEK_SET);
	MD2->skins=malloc(sizeof(MD2_Skin_t)*MD2->header.numSkins);
	for(i=0;i<MD2->header.numSkins;i++)
		fread(&MD2->skins[i], sizeof(MD2_Skin_t), 1, stream);

	fseek(stream, MD2->header.offsetTexCoords, SEEK_SET);
	MD2->texcoords=malloc(sizeof(MD2_TextureCoordinate_t)*MD2->header.numTexCoords);
	for(i=0;i<MD2->header.numTexCoords;i++)
		fread(&MD2->texcoords[i], sizeof(MD2_TextureCoordinate_t), 1, stream);

	fseek(stream, MD2->header.offsetTriangles, SEEK_SET);
	MD2->triangles=malloc(sizeof(MD2_Triangle_t)*MD2->header.numTriangles);
	for(i=0;i<MD2->header.numTriangles;i++)
		fread(&MD2->triangles[i], sizeof(MD2_Triangle_t), 1, stream);

	fseek(stream, MD2->header.offsetFrames, SEEK_SET);
	MD2->frames=malloc(sizeof(MD2_Frame_t)*MD2->header.numFrames);
	for(i=0;i<MD2->header.numFrames;i++)
	{
		MD2_Alias_Frame_t *frame=(MD2_Alias_Frame_t *)buffer;

		MD2->frames[i].vertices=malloc(sizeof(MD2_TriangleVertex_t)*MD2->header.numVertices);

		fread(frame, 1, MD2->header.frameSize, stream);

		strcpy(MD2->frames[i].name, frame->name);

		for(j=0;j<MD2->header.numVertices;j++)
		{
			MD2->frames[i].vertices[j].vertex[0]=(float)frame->alias_vertices[j].vertex[0]*frame->scale[0]+frame->translate[0];
			MD2->frames[i].vertices[j].vertex[1]=(float)frame->alias_vertices[j].vertex[1]*frame->scale[1]+frame->translate[1];
			MD2->frames[i].vertices[j].vertex[2]=(float)frame->alias_vertices[j].vertex[2]*frame->scale[2]+frame->translate[2];
		}
	}

	fseek(stream, MD2->header.offsetGlCommands, SEEK_SET);
	MD2->glcommandbuffer=malloc(sizeof(int)*MD2->header.numGlCommands);
	fread(MD2->glcommandbuffer, sizeof(int), MD2->header.numGlCommands, stream);

	fclose(stream);

	for(i=0;i<MD2->header.numFrames;i++)
	{
		for(j=0;j<MD2->header.numVertices;j++)
		{
			MD2->frames[i].vertices[j].normal[0]=0.0f;
			MD2->frames[i].vertices[j].normal[1]=0.0f;
			MD2->frames[i].vertices[j].normal[2]=0.0f;
		}

		for(j=0;j<MD2->header.numTriangles;j++)
		{
			float n[3], v0[3], v1[3];

			v0[0]=MD2->frames[i].vertices[MD2->triangles[j].vertexIndices[1]].vertex[0]-MD2->frames[i].vertices[MD2->triangles[j].vertexIndices[0]].vertex[0];
			v0[1]=MD2->frames[i].vertices[MD2->triangles[j].vertexIndices[1]].vertex[1]-MD2->frames[i].vertices[MD2->triangles[j].vertexIndices[0]].vertex[1];
			v0[2]=MD2->frames[i].vertices[MD2->triangles[j].vertexIndices[1]].vertex[2]-MD2->frames[i].vertices[MD2->triangles[j].vertexIndices[0]].vertex[2];
			v1[0]=MD2->frames[i].vertices[MD2->triangles[j].vertexIndices[2]].vertex[0]-MD2->frames[i].vertices[MD2->triangles[j].vertexIndices[0]].vertex[0];
			v1[1]=MD2->frames[i].vertices[MD2->triangles[j].vertexIndices[2]].vertex[1]-MD2->frames[i].vertices[MD2->triangles[j].vertexIndices[0]].vertex[1];
			v1[2]=MD2->frames[i].vertices[MD2->triangles[j].vertexIndices[2]].vertex[2]-MD2->frames[i].vertices[MD2->triangles[j].vertexIndices[0]].vertex[2];

			n[0]=(v0[1]*v1[2]-v0[2]*v1[1]);
			n[1]=(v0[2]*v1[0]-v0[0]*v1[2]);
			n[2]=(v0[0]*v1[1]-v0[1]*v1[0]);

			for(k=0;k<3;k++)
			{
				MD2->frames[i].vertices[MD2->triangles[j].vertexIndices[k]].normal[0]=-n[0];
				MD2->frames[i].vertices[MD2->triangles[j].vertexIndices[k]].normal[2]=-n[2];
				MD2->frames[i].vertices[MD2->triangles[j].vertexIndices[k]].normal[1]=-n[1];
			}
		}

		for(j=0;j<MD2->header.numVertices;j++)
		{
			float len=(float)sqrt((MD2->frames[i].vertices[j].normal[0]*MD2->frames[i].vertices[j].normal[0])+(MD2->frames[i].vertices[j].normal[1]*MD2->frames[i].vertices[j].normal[1])+(MD2->frames[i].vertices[j].normal[2]*MD2->frames[i].vertices[j].normal[2]));

			if(len!=0.0f)
			{
				MD2->frames[i].vertices[j].normal[0]/=len;
				MD2->frames[i].vertices[j].normal[1]/=len;
				MD2->frames[i].vertices[j].normal[2]/=len;
			}
		}
	}

	for(i=0;i<MD2->header.numFrames;i++)
	{
		for(j=0;j<MD2->header.numTriangles;j++)
		{
			float v0[3], v1[3], v2[3], t0[2], t1[2], t2[2], tangent[3], binormal[3];

			v0[0]=MD2->frames[i].vertices[MD2->triangles[j].vertexIndices[0]].vertex[0];
			v0[1]=MD2->frames[i].vertices[MD2->triangles[j].vertexIndices[0]].vertex[1];
			v0[2]=MD2->frames[i].vertices[MD2->triangles[j].vertexIndices[0]].vertex[2];

			v1[0]=MD2->frames[i].vertices[MD2->triangles[j].vertexIndices[1]].vertex[0];
			v1[1]=MD2->frames[i].vertices[MD2->triangles[j].vertexIndices[1]].vertex[1];
			v1[2]=MD2->frames[i].vertices[MD2->triangles[j].vertexIndices[1]].vertex[2];

			v2[0]=MD2->frames[i].vertices[MD2->triangles[j].vertexIndices[2]].vertex[0];
			v2[1]=MD2->frames[i].vertices[MD2->triangles[j].vertexIndices[2]].vertex[1];
			v2[2]=MD2->frames[i].vertices[MD2->triangles[j].vertexIndices[2]].vertex[2];

			t0[0]=(float)MD2->texcoords[MD2->triangles[j].textureIndices[0]].s/MD2->header.skinWidth;
			t0[1]=(float)MD2->texcoords[MD2->triangles[j].textureIndices[0]].t/MD2->header.skinHeight;

			t1[0]=(float)MD2->texcoords[MD2->triangles[j].textureIndices[1]].s/MD2->header.skinWidth;
			t1[1]=(float)MD2->texcoords[MD2->triangles[j].textureIndices[1]].t/MD2->header.skinHeight;

			t2[0]=(float)MD2->texcoords[MD2->triangles[j].textureIndices[2]].s/MD2->header.skinWidth;
			t2[1]=(float)MD2->texcoords[MD2->triangles[j].textureIndices[2]].t/MD2->header.skinHeight;

			tangent_basis(v0, v1, v2, t0, t1, t2, tangent, binormal);

			for(k=0;k<3;k++)
			{
				MD2->frames[i].vertices[MD2->triangles[j].vertexIndices[k]].tangent[0]=tangent[0];
				MD2->frames[i].vertices[MD2->triangles[j].vertexIndices[k]].tangent[1]=tangent[1];
				MD2->frames[i].vertices[MD2->triangles[j].vertexIndices[k]].tangent[2]=tangent[2];

				MD2->frames[i].vertices[MD2->triangles[j].vertexIndices[k]].binormal[0]=binormal[0];
				MD2->frames[i].vertices[MD2->triangles[j].vertexIndices[k]].binormal[1]=binormal[1];
				MD2->frames[i].vertices[MD2->triangles[j].vertexIndices[k]].binormal[2]=binormal[2];
			}
		}

		for(j=0;j<MD2->header.numVertices;j++)
		{
			float len;

			len=(float)sqrt((MD2->frames[i].vertices[j].tangent[0]*MD2->frames[i].vertices[j].tangent[0])+(MD2->frames[i].vertices[j].tangent[1]*MD2->frames[i].vertices[j].tangent[1])+(MD2->frames[i].vertices[j].tangent[2]*MD2->frames[i].vertices[j].tangent[2]));
			if(len!=0.0f)
			{
				MD2->frames[i].vertices[j].tangent[0]/=len;
				MD2->frames[i].vertices[j].tangent[1]/=len;
				MD2->frames[i].vertices[j].tangent[2]/=len;
			}

			len=(float)sqrt((MD2->frames[i].vertices[j].binormal[0]*MD2->frames[i].vertices[j].binormal[0])+(MD2->frames[i].vertices[j].binormal[1]*MD2->frames[i].vertices[j].binormal[1])+(MD2->frames[i].vertices[j].binormal[2]*MD2->frames[i].vertices[j].binormal[2]));
			if(len!=0.0f)
			{
				MD2->frames[i].vertices[j].binormal[0]/=len;
				MD2->frames[i].vertices[j].binormal[1]/=len;
				MD2->frames[i].vertices[j].binormal[2]/=len;
			}
		}
	}

	return TRUE;
}

void MD2_DrawFrameDOT3(MD2_Model_t *MD2, int Frame, float lightpos[3])
{
	int i, j;

	glBegin(GL_TRIANGLES);
	for(i=0;i<MD2->header.numTriangles;i++)
	{
		for(j=0;j<3;j++)
		{
			float m[9], lv[3], tc[3], len;

			lv[0]=(lightpos[0]-MD2->frames[Frame].vertices[MD2->triangles[i].vertexIndices[j]].vertex[0]);
			lv[1]=(lightpos[1]-MD2->frames[Frame].vertices[MD2->triangles[i].vertexIndices[j]].vertex[1]);
			lv[2]=(lightpos[2]-MD2->frames[Frame].vertices[MD2->triangles[i].vertexIndices[j]].vertex[2]);

			len=(float)sqrt((lv[0]*lv[0])+(lv[1]*lv[1])+(lv[2]*lv[2]));
			if(len!=0.0f)
			{
				lv[0]/=len;
				lv[1]/=len;
				lv[2]/=len;
			}

			m[0]=MD2->frames[Frame].vertices[MD2->triangles[i].vertexIndices[j]].tangent[0];
			m[1]=MD2->frames[Frame].vertices[MD2->triangles[i].vertexIndices[j]].binormal[0];
			m[2]=MD2->frames[Frame].vertices[MD2->triangles[i].vertexIndices[j]].normal[0];
			m[3]=MD2->frames[Frame].vertices[MD2->triangles[i].vertexIndices[j]].tangent[1];
			m[4]=MD2->frames[Frame].vertices[MD2->triangles[i].vertexIndices[j]].binormal[1];
			m[5]=MD2->frames[Frame].vertices[MD2->triangles[i].vertexIndices[j]].normal[1];
			m[6]=MD2->frames[Frame].vertices[MD2->triangles[i].vertexIndices[j]].tangent[2];
			m[7]=MD2->frames[Frame].vertices[MD2->triangles[i].vertexIndices[j]].binormal[2];
			m[8]=MD2->frames[Frame].vertices[MD2->triangles[i].vertexIndices[j]].normal[2];

			tc[0]=(lv[0]*m[0])+(lv[1]*m[3])+(lv[2]*m[6]);
			tc[1]=(lv[0]*m[1])+(lv[1]*m[4])+(lv[2]*m[7]);
			tc[2]=(lv[0]*m[2])+(lv[1]*m[5])+(lv[2]*m[8]);

			tc[0]=(tc[0]*0.5f)+0.5f;
			tc[1]=(tc[1]*0.5f)+0.5f;
			tc[2]=(tc[2]*0.5f)+0.5f;

			glMultiTexCoord2fARB(GL_TEXTURE0_ARB, (float)MD2->texcoords[MD2->triangles[i].textureIndices[j]].s/MD2->header.skinWidth, -(float)MD2->texcoords[MD2->triangles[i].textureIndices[j]].t/MD2->header.skinHeight);
			glMultiTexCoord2fARB(GL_TEXTURE1_ARB, (float)MD2->texcoords[MD2->triangles[i].textureIndices[j]].s/MD2->header.skinWidth, -(float)MD2->texcoords[MD2->triangles[i].textureIndices[j]].t/MD2->header.skinHeight);
			glColor3f(tc[0], tc[1], tc[2]);
			glNormal3f(MD2->frames[Frame].vertices[MD2->triangles[i].vertexIndices[j]].normal[0], MD2->frames[Frame].vertices[MD2->triangles[i].vertexIndices[j]].normal[1], MD2->frames[Frame].vertices[MD2->triangles[i].vertexIndices[j]].normal[2]);
			glVertex3f(MD2->frames[Frame].vertices[MD2->triangles[i].vertexIndices[j]].vertex[0], MD2->frames[Frame].vertices[MD2->triangles[i].vertexIndices[j]].vertex[1], MD2->frames[Frame].vertices[MD2->triangles[i].vertexIndices[j]].vertex[2]);
		}
	}
	glEnd();
}

void MD2_DrawFrame(MD2_Model_t *MD2, int Frame)
{
	int i, j;

	glBegin(GL_TRIANGLES);
	for(i=0;i<MD2->header.numTriangles;i++)
	{
		for(j=0;j<3;j++)
		{
			glMultiTexCoord2fARB(GL_TEXTURE0_ARB, (float)MD2->texcoords[MD2->triangles[i].textureIndices[j]].s/MD2->header.skinWidth, -(float)MD2->texcoords[MD2->triangles[i].textureIndices[j]].t/MD2->header.skinHeight);
			glMultiTexCoord2fARB(GL_TEXTURE1_ARB, (float)MD2->texcoords[MD2->triangles[i].textureIndices[j]].s/MD2->header.skinWidth, -(float)MD2->texcoords[MD2->triangles[i].textureIndices[j]].t/MD2->header.skinHeight);
			glNormal3f(MD2->frames[Frame].vertices[MD2->triangles[i].vertexIndices[j]].normal[0], MD2->frames[Frame].vertices[MD2->triangles[i].vertexIndices[j]].normal[1], MD2->frames[Frame].vertices[MD2->triangles[i].vertexIndices[j]].normal[2]);
			glVertex3f(MD2->frames[Frame].vertices[MD2->triangles[i].vertexIndices[j]].vertex[0], MD2->frames[Frame].vertices[MD2->triangles[i].vertexIndices[j]].vertex[1], MD2->frames[Frame].vertices[MD2->triangles[i].vertexIndices[j]].vertex[2]);
		}
	}
	glEnd();
}

void MD2_DrawVectors(MD2_Model_t *MD2, int Frame)
{
	int i, j;

	glBegin(GL_LINES);
	for(i=0;i<MD2->header.numTriangles;i++)
	{
		for(j=0;j<3;j++)
		{
			glColor3f(1.0f, 0.0f, 0.0f);
			glVertex3f(MD2->frames[Frame].vertices[MD2->triangles[i].vertexIndices[j]].vertex[0], MD2->frames[Frame].vertices[MD2->triangles[i].vertexIndices[j]].vertex[1], MD2->frames[Frame].vertices[MD2->triangles[i].vertexIndices[j]].vertex[2]);
			glVertex3f(MD2->frames[Frame].vertices[MD2->triangles[i].vertexIndices[j]].vertex[0]+MD2->frames[Frame].vertices[MD2->triangles[i].vertexIndices[j]].normal[0], MD2->frames[Frame].vertices[MD2->triangles[i].vertexIndices[j]].vertex[1]+MD2->frames[Frame].vertices[MD2->triangles[i].vertexIndices[j]].normal[1], MD2->frames[Frame].vertices[MD2->triangles[i].vertexIndices[j]].vertex[2]+MD2->frames[Frame].vertices[MD2->triangles[i].vertexIndices[j]].normal[2]);

			glColor3f(0.0f, 1.0f, 0.0f);
			glVertex3f(MD2->frames[Frame].vertices[MD2->triangles[i].vertexIndices[j]].vertex[0], MD2->frames[Frame].vertices[MD2->triangles[i].vertexIndices[j]].vertex[1], MD2->frames[Frame].vertices[MD2->triangles[i].vertexIndices[j]].vertex[2]);
			glVertex3f(MD2->frames[Frame].vertices[MD2->triangles[i].vertexIndices[j]].vertex[0]+MD2->frames[Frame].vertices[MD2->triangles[i].vertexIndices[j]].tangent[0], MD2->frames[Frame].vertices[MD2->triangles[i].vertexIndices[j]].vertex[1]+MD2->frames[Frame].vertices[MD2->triangles[i].vertexIndices[j]].tangent[1], MD2->frames[Frame].vertices[MD2->triangles[i].vertexIndices[j]].vertex[2]+MD2->frames[Frame].vertices[MD2->triangles[i].vertexIndices[j]].tangent[2]);

			glColor3f(0.0f, 0.0f, 1.0f);
			glVertex3f(MD2->frames[Frame].vertices[MD2->triangles[i].vertexIndices[j]].vertex[0], MD2->frames[Frame].vertices[MD2->triangles[i].vertexIndices[j]].vertex[1], MD2->frames[Frame].vertices[MD2->triangles[i].vertexIndices[j]].vertex[2]);
			glVertex3f(MD2->frames[Frame].vertices[MD2->triangles[i].vertexIndices[j]].vertex[0]+MD2->frames[Frame].vertices[MD2->triangles[i].vertexIndices[j]].binormal[0], MD2->frames[Frame].vertices[MD2->triangles[i].vertexIndices[j]].vertex[1]+MD2->frames[Frame].vertices[MD2->triangles[i].vertexIndices[j]].binormal[1], MD2->frames[Frame].vertices[MD2->triangles[i].vertexIndices[j]].vertex[2]+MD2->frames[Frame].vertices[MD2->triangles[i].vertexIndices[j]].binormal[2]);
		}

	}
	glEnd();
}

void MD2_DrawFramei(MD2_Model_t *MD2, int Frame1, int Frame2, float Pol)
{
	int i=0, val=MD2->glcommandbuffer[i++];
	float x1, y1, z1, x2, y2, z2, nx1, ny1, nz1, nx2, ny2, nz2;

	while(val!=0)
	{
		int count;

		if(val>0)
		{
			glBegin(GL_TRIANGLE_STRIP);

			count=val;
		}
		else
		{
			glBegin(GL_TRIANGLE_FAN);

			count=-val;
		}

		while(count--)
		{
			float s=*(float *)&MD2->glcommandbuffer[i++];
			float t=*(float *)&MD2->glcommandbuffer[i++];
			int index=MD2->glcommandbuffer[i++];

			x1=MD2->frames[Frame1].vertices[index].vertex[0];
			y1=MD2->frames[Frame1].vertices[index].vertex[1];
			z1=MD2->frames[Frame1].vertices[index].vertex[2];

			x2=MD2->frames[Frame2].vertices[index].vertex[0];
			y2=MD2->frames[Frame2].vertices[index].vertex[1];
			z2=MD2->frames[Frame2].vertices[index].vertex[2];

			nx1=MD2->frames[Frame1].vertices[index].normal[0];
			ny1=MD2->frames[Frame1].vertices[index].normal[1];
			nz1=MD2->frames[Frame1].vertices[index].normal[2];

			nx2=MD2->frames[Frame2].vertices[index].normal[0];
			ny2=MD2->frames[Frame2].vertices[index].normal[1];
			nz2=MD2->frames[Frame2].vertices[index].normal[2];

			glTexCoord2f(s, -t);
			glNormal3f(nx1+Pol*(nx2-nx1), ny1+Pol*(ny2-ny1), nz1+Pol*(nz2-nz1));
			glVertex3f(x1+Pol*(x2-x1), y1+Pol*(y2-y1), z1+Pol*(z2-z1));
		}

		glEnd();

		val=MD2->glcommandbuffer[i++];
	}
}
